/***********************************************************************\
 *                                                                     * 
 *                         PROJECT   UniKalow                            *
 *                                                                     * 
 *---------------------------------------------------------------------*
 *                                By                                   *
 *                                                                     *
 *                            Zemin Ning                               *
 *                                                                     *
 *          Copyright (C) 2010-2011 by Genome Research Limited         *
 *                       All rights reserved                           *
 *                                                                     *
 *---------------------------------------------------------------------*
 #######################################################################
 #                                                                     #
 #             <------   LICENCE NOTICE   ------>                      #
 #                                                                     #
 # This is a licensed software developed by Genome Research Limited    #
 # (GRL) for genomic sequence assembling. For both commercial and non- # 
 # commercial purposes, a licence must be obtained from GRL before     #
 # use. Under no circumstances, the users should copy, modify and      #
 # redistribut the software as a whole or even part of it without      #
 # permission from GRL. For more information about the software and    #
 # its ducumentation particularly, please contact either one of the    # 
 # authors or GRL.                                                     #
 #######################################################################
 *---------------------------------------------------------------------*/

/****************************************************************************/

 
#include <math.h>
#include <values.h>
#include <stdio.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include "fasta.h"

#define GT '>'
#define GT4 (((((GT<<8)+GT)<<8)+GT)<<8)+GT

#define Max_N_NameBase 400
#define Max_N_NameBase2 400
static char workdir[1000];
static char bindir[1000];
static char tmpdir[1000];

/* SSAS default parameters   */
static int IMOD=0;
static int num_uniq= 10;
static int n_thread= 4;
static int num_depth= 10;
static int num_copy= 100000;
static int kmer_size = 41;
static int trim_len = 80;
static int screen_flag = 1;

int main(int argc, char **argv)
{
    int i,nSeq,args,num_libs;
    char **cmatrix(B64_long nrl,B64_long nrh,B64_long ncl,B64_long nch);
    FILE *namef,*fpconfig;
    char *st,*ed,**rdname1,**rdname2,**rdfiles;
    char allfiles[5000],line[2000],tempc[1000],syscmd[2000];
    pid_t pid;

    if(argc < 2)
    {
      printf("Usage: %s <-kmer 41> <-thread 4> <-trim 80> <-uniq 10> <-depth 100000> <config_file> <assembly_file>\n",argv[0]);

      exit(1);
    }

    nSeq=0;
    args=1;
    for(i=1;i<argc;i++)
    {
       if(!strcmp(argv[i],"-mod"))
       {
         sscanf(argv[++i],"%d",&IMOD); 
         args=args+2;
       }
       else if(!strcmp(argv[i],"-kmer"))
       {
         sscanf(argv[++i],"%d",&kmer_size);
         args=args+2;
       }
       else if(!strcmp(argv[i],"-uniq"))
       {
         sscanf(argv[++i],"%d",&num_uniq);
         args=args+2;
       }
       else if(!strcmp(argv[i],"-copy"))
       {
         sscanf(argv[++i],"%d",&num_copy);
         args=args+2;
       }
       else if(!strcmp(argv[i],"-thread"))
       {
         sscanf(argv[++i],"%d",&n_thread);
         args=args+2;
       }
       else if(!strcmp(argv[i],"-depth"))
       {
         sscanf(argv[++i],"%d",&num_depth);
         args=args+2;
       }
       else if(!strcmp(argv[i],"-screen"))
       {
         sscanf(argv[++i],"%d",&screen_flag);
         args=args+2;
       }
       else if(!strcmp(argv[i],"-trim"))
       {
         sscanf(argv[++i],"%d",&trim_len);
         args=args+2;
       }
    }

    pid = getpid();

    rdname1 = cmatrix(0,1000,0,Max_N_NameBase);
    rdname2 = cmatrix(0,1000,0,Max_N_NameBase);
    rdfiles = cmatrix(0,1000,0,Max_N_NameBase);
    memset(tempc,'\0',1000);
    if (!getcwd(tempc, sizeof(tempc)))
    {
      exit(1);
    } 
    memset(tmpdir,'\0',1000);
    memset(workdir,'\0',1000);
    sprintf(tmpdir,"%s/",tempc);
    sprintf(workdir,"%s/tmp.rununik.%d/",tempc,pid);
//    printf("%s\n",workdir);

    st = argv[0];
    ed = strrchr(argv[0],'/');
    memset(tempc,'\0',1000);
    strncpy(tempc,argv[0],ed-st);
    memset(bindir,'\0',1000);
    sprintf(bindir,"%s/bin",tempc);

    memset(syscmd,'\0',2000);
    sprintf(syscmd,"mkdir %s",workdir);
    system (syscmd);

    memset(syscmd,'\0',2000);
    sprintf(syscmd,"cp %s %s > /dev/null",argv[args],workdir);
    system (syscmd);

    chdir(workdir);

   if(screen_flag)
   {
    if((namef = fopen(argv[args],"r")) == NULL)
    {
      printf("ERROR main:: reads group file \n");
      exit(1);
    }

/*  read the congfig file         */
    num_libs = 0;

    while(!feof(namef))
    {
       int j,r_len,n_len;

       fgets(line,2000,namef);
       if(feof(namef)) break;
       r_len = strlen(line) - 1;
       if((strncmp(line,"q1=",3))==0)
       {
         st = line;
         ed = strrchr(line,'/');
	 if(ed==NULL)
           n_len = 3;
	 else
           n_len = ed-st+1;
         memset(rdname1[num_libs],'\0',Max_N_NameBase);
         for(j=n_len;j<r_len;j++)
	    rdname1[num_libs][j-n_len] = line[j];
       }
       else if((strncmp(line,"q2=",3))==0)    
       {
         st = line;
         ed = strrchr(line,'/');
	 if(ed==NULL)
           n_len = 3;
	 else
           n_len = ed-st+1;
         memset(rdname2[num_libs],'\0',Max_N_NameBase);
         for(j=n_len;j<r_len;j++)
	    rdname2[num_libs][j-n_len] = line[j];
         num_libs++; 
       }
       else if((strncmp(line,"f1=",3))==0)    
       {
         st = line;
         ed = strrchr(line,'/');
	 if(ed==NULL)
           n_len = 3;
	 else
           n_len = ed-st+1;
         memset(rdname1[num_libs],'\0',Max_N_NameBase);
         for(j=n_len;j<r_len;j++)
	    rdname1[num_libs][j-n_len] = line[j];
       }
       else if((strncmp(line,"f2=",3))==0)    
       {
         st = line;
         ed = strrchr(line,'/');
	 if(ed==NULL)
           n_len = 3;
	 else
           n_len = ed-st+1;
         memset(rdname2[num_libs],'\0',Max_N_NameBase);
         for(j=n_len;j<r_len;j++)
	    rdname2[num_libs][j-n_len] = line[j];
         num_libs++; 
       }
       else if((strncmp(line,"q=",2))==0)    
       {
       }
       else if((strncmp(line,"f=",2))==0)    
       {
       }
    }

/*  merge, trim and hash the read files   */
    memset(allfiles,'\0',5000);
    for(i=0;i<num_libs;i++)
    {
       sprintf(rdfiles[i],"%s%04d","readfile_",i);
       memset(syscmd,'\0',2000);
       sprintf(syscmd,"%s/ssaha_merge -file 2 -trim %d %s%s %s%s %s_trim > /dev/null",bindir,trim_len,tmpdir,rdname1[i],tmpdir,rdname2[i],rdfiles[i]);
       system (syscmd);
 
/*       memset(syscmd,'\0',2000);
       sprintf(syscmd,"%s/ssaha_trimm -file 4 -trim %d %s %s_trim > /dev/null",bindir,trim_len,rdfiles[i],rdfiles[i]); 
       system (syscmd);   */
 
       memset(syscmd,'\0',2000);
       sprintf(syscmd,"%s/Kmer_Screen hash -kmer 31 -thread %d %s_trim > /dev/null",bindir,n_thread,rdfiles[i]);
       system (syscmd);

       strcat(allfiles,rdfiles[i]);
       strcat(allfiles,"_trim ");
 
       memset(syscmd,'\0',2000);
       sprintf(syscmd,"rm -rf %s ",rdfiles[i]); 
       system (syscmd);
    }

    memset(syscmd,'\0',2000);
    sprintf(syscmd,"%s/Kmer_Screen sort -kmer 31 -thread %d %s > /dev/null",bindir,n_thread,allfiles);
    system (syscmd);

    memset(syscmd,'\0',2000);
    sprintf(syscmd,"%s/Kmer_Screen screen -kmer 31 -thread %d -uniq %d -copy %d -cut %d %s > /dev/null",bindir,n_thread,num_uniq,num_copy,num_depth,allfiles);
    system (syscmd);

    memset(syscmd,'\0',2000);
    sprintf(syscmd,"rm -rf sorted.* ");
    system (syscmd);

    for(i=0;i<num_libs;i++)
    {
       memset(syscmd,'\0',2000);
       sprintf(syscmd,"%s/ssaha_reads -file 23 %s_trim.screen %s.screen %s.screen > /dev/null",bindir,rdfiles[i],rdname1[i],rdname2[i]);
       system (syscmd);

       memset(syscmd,'\0',2000);
       sprintf(syscmd,"rm -rf %s_trim* ",rdfiles[i]);
       system (syscmd);
    }

    if((fpconfig = fopen("assembly_config","w")) == NULL)
    {
       printf("ERROR main:: reads group file \n");
       exit(1);
    }
    if((namef = fopen(argv[args],"r")) == NULL)
    {
      printf("ERROR main:: reads group file \n");
      exit(1);
    }

/*  read the congfig file         */
    num_libs = 0;

    while(!feof(namef))
    {
       fgets(line,2000,namef);
       if(feof(namef)) break;
       if((strncmp(line,"q1=",3))==0)
       {
         fprintf(fpconfig,"q1=%s%s.screen\n",workdir,rdname1[num_libs]);
       }
       else if((strncmp(line,"q2=",3))==0)    
       {
         fprintf(fpconfig,"q2=%s%s.screen\n",workdir,rdname2[num_libs]);
         num_libs++; 
       }
       else if((strncmp(line,"f1=",3))==0)    
       {
         fprintf(fpconfig,"f1=%s%s.screen\n",workdir,rdname1[num_libs]);
       }
       else if((strncmp(line,"f2=",3))==0)    
       {
         fprintf(fpconfig,"f2=%s%s.screen\n",workdir,rdname2[num_libs]);
         num_libs++; 
       }
       else if((strncmp(line,"q=",2))==0)    
       {
       }
       else if((strncmp(line,"f=",2))==0)    
       {
       }
       else
         fprintf(fpconfig,"%s",line);
    }
    fclose(fpconfig);
    fclose(namef);


    memset(syscmd,'\0',2000);
    if(kmer_size>=63)
      sprintf(syscmd,"%s/SOAPdenovo-127mer all -s %s -K %d -o kmer_%03d > /dev/null",bindir,"assembly_config",kmer_size,kmer_size); 
    else if(kmer_size>=32)
      sprintf(syscmd,"%s/SOAPdenovo-63mer all -s %s -K %d -o kmer_%03d > /dev/null",bindir,"assembly_config",kmer_size,kmer_size); 
    else
      sprintf(syscmd,"%s/SOAPdenovo-31mer all -s %s -K %d -o kmer_%03d > /dev/null",bindir,"assembly_config",kmer_size,kmer_size); 
    system (syscmd);

    memset(syscmd,'\0',2000);
    sprintf(syscmd,"%s/GapCloser -o kmer_%03d.gapfill -b %s -a kmer_%03d.scafSeq > /dev/null",bindir,kmer_size,"assembly_config",kmer_size); 
    system (syscmd);

   }
   else
   {
    memset(syscmd,'\0',2000);
    if(kmer_size>=63)
      sprintf(syscmd,"%s/SOAPdenovo-127mer all -s %s -K %d -o kmer_%03d > /dev/null",bindir,argv[args],kmer_size,kmer_size);
    else if(kmer_size>=32)
      sprintf(syscmd,"%s/SOAPdenovo-63mer all -s %s -K %d -o kmer_%03d > /dev/null",bindir,argv[args],kmer_size,kmer_size);
    else
      sprintf(syscmd,"%s/SOAPdenovo-31mer all -s %s -K %d -o kmer_%03d > /dev/null",bindir,argv[args],kmer_size,kmer_size); 
    system (syscmd);

    memset(syscmd,'\0',2000);
    sprintf(syscmd,"%s/GapCloser -o kmer_%03d.gapfill -b %s -a kmer_%03d.scafSeq > /dev/null",bindir,kmer_size,argv[args],kmer_size); 
    system (syscmd);
   }
    memset(syscmd,'\0',2000);
    sprintf(syscmd,"mv kmer_%03d.gapfill %s%s.super.fasta ",kmer_size,tmpdir,argv[args+1]); 
    system (syscmd);

    memset(syscmd,'\0',2000);
    sprintf(syscmd,"rm -rf kmer_%03d.* ",kmer_size); 
    system (syscmd);

    printf("Job finished!\n");
    return EXIT_SUCCESS;

}
/* end of the main */

#define SWAP(a,b) temp=(a);(a)=b;(b)=temp;

/*   Subroutine to sort an array arr[0,...,n-1] into ascending order while
     making the corresponding reaarangement of the array brr[0,...,n-1]
     by the use of Quicksort (Sedgwick, R. 1978, Communications o fthe ACM,
     vol. 21, pp. 847-857) also see Numerical Recipes in C                  */  

/* =============================== */
void ArraySort_Long(int n, B64_long *arr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,NSTACK=50,istack[NSTACK];
     B64_long a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
             }
             arr[i+1]=a;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}


/* =============================== */
void ArraySort_Int(int n, int *arr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,NSTACK=50,istack[NSTACK];
     int a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
             }
             arr[i+1]=a;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}


/* =============================== */
void ArraySort_Mix(int n, B64_long *arr, int *brr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     B64_long a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
                brr[i+1]=brr[i];
             }
             arr[i+1]=a;
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);
          SWAP(brr[k],brr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
            SWAP(brr[m],brr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
            SWAP(brr[m+1],brr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          b=brr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
             SWAP(brr[i],brr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}

/* =============================== */
void ArraySort_Int2(int n, int *arr, int *brr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     int a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
                brr[i+1]=brr[i];
             }
             arr[i+1]=a;
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);
          SWAP(brr[k],brr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
            SWAP(brr[m],brr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
            SWAP(brr[m+1],brr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          b=brr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
             SWAP(brr[i],brr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}

/*   function to sort an array into a decreasing order:  a>b>c>....    */  
/* =============================== */
void ArraySort2_Int2(int n, int *arr, int *brr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     int a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]>=a) break;
                arr[i+1]=arr[i];
                brr[i+1]=brr[i];
             }
             arr[i+1]=a;
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);
          SWAP(brr[k],brr[m+1]);

          if(arr[m]<arr[ir])
          {
            SWAP(arr[m],arr[ir]);
            SWAP(brr[m],brr[ir]);
          }

          if(arr[m+1]<arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
            SWAP(brr[m+1],brr[ir]);
          }

          if(arr[m]<arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          b=brr[m+1];
          for(;;)
          {
             do i++; while (arr[i]>a);
             do j--; while (arr[j]<a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
             SWAP(brr[i],brr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}

/* =============================== */
void ArraySort_Mix3(int n, B64_long *arr, int *brr, int *crr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,c,NSTACK=50,istack[NSTACK];
     B64_long a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             b=brr[j];
             c=crr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
                brr[i+1]=brr[i];
                crr[i+1]=crr[i];
             }
             arr[i+1]=a;
             brr[i+1]=b;
             crr[i+1]=c;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);
          SWAP(brr[k],brr[m+1]);
          SWAP(crr[k],crr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
            SWAP(brr[m],brr[ir]);
            SWAP(crr[m],crr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
            SWAP(brr[m+1],brr[ir]);
            SWAP(crr[m+1],crr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
            SWAP(brr[m],brr[m+1]);
            SWAP(crr[m],crr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          b=brr[m+1];
          c=crr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
             SWAP(brr[i],brr[j]);
             SWAP(crr[i],crr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          brr[m+1]=brr[j];
          brr[j]=b;
          crr[m+1]=crr[j];
          crr[j]=c;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}


/*   to swap the string arrays           */
/* ============================================= */
void s_swap(char **Pair_Name, int i, int j)
/* ============================================= */
{
     char temp[Max_N_NameBase];

     strcpy(temp,Pair_Name[i]);
     strcpy(Pair_Name[i],Pair_Name[j]);
     strcpy(Pair_Name[j],temp);
}


/*   to sort the string array in order          */
/* ============================================= */
void ArraySort_String(int n, char **Pair_Name, int *brr)
/* ============================================= */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     int temp,MIN=7;
     char p[Max_N_NameBase];

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             strcpy(p,Pair_Name[j]);
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(strcmp(Pair_Name[i],p)<=0) break;
                strcpy(Pair_Name[i+1],Pair_Name[i]);
                brr[i+1]=brr[i];
             }
             strcpy(Pair_Name[i+1],p);
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          s_swap(Pair_Name,k,m+1);
          SWAP(brr[k],brr[m+1]);

          if(strcmp(Pair_Name[m],Pair_Name[ir])>0)
          {
            s_swap(Pair_Name,m,ir);
            SWAP(brr[m],brr[ir]);
          }

          if(strcmp(Pair_Name[m+1],Pair_Name[ir])>0)
          {
            s_swap(Pair_Name,m+1,ir);
            SWAP(brr[m+1],brr[ir]);
          }

          if(strcmp(Pair_Name[m],Pair_Name[m+1])>0)
          {
            s_swap(Pair_Name,m,m+1);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          strcpy(p,Pair_Name[m+1]);
          b=brr[m+1];
          for(;;)
          {
             do i++; while (strcmp(Pair_Name[i],p)<0);
             do j--; while (strcmp(Pair_Name[j],p)>0);
             if(j<i) break;
             s_swap(Pair_Name,i,j);
             SWAP(brr[i],brr[j]);
          }
          strcpy(Pair_Name[m+1],Pair_Name[j]);
          strcpy(Pair_Name[j],p);
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}

/*   to swap the string arrays           */
/* ============================================= */
void s_swap2(char **Pair_Name, int i, int j)
/* ============================================= */
{
     char temp[Max_N_NameBase2];

     strcpy(temp,Pair_Name[i]);
     strcpy(Pair_Name[i],Pair_Name[j]);
     strcpy(Pair_Name[j],temp);
}


/*   to sort the string array in order          */
/* ============================================= */
void ArraySort_String2(int n, char **Pair_Name, int *brr)
/* ============================================= */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     int temp,MIN=7;
     char p[Max_N_NameBase2];

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             strcpy(p,Pair_Name[j]);
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(strcmp(Pair_Name[i],p)<=0) break;
                strcpy(Pair_Name[i+1],Pair_Name[i]);
                brr[i+1]=brr[i];
             }
             strcpy(Pair_Name[i+1],p);
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          s_swap2(Pair_Name,k,m+1);
          SWAP(brr[k],brr[m+1]);

          if(strcmp(Pair_Name[m],Pair_Name[ir])>0)
          {
            s_swap2(Pair_Name,m,ir);
            SWAP(brr[m],brr[ir]);
          }

          if(strcmp(Pair_Name[m+1],Pair_Name[ir])>0)
          {
            s_swap2(Pair_Name,m+1,ir);
            SWAP(brr[m+1],brr[ir]);
          }

          if(strcmp(Pair_Name[m],Pair_Name[m+1])>0)
          {
            s_swap2(Pair_Name,m,m+1);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          strcpy(p,Pair_Name[m+1]);
          b=brr[m+1];
          for(;;)
          {
             do i++; while (strcmp(Pair_Name[i],p)<0);
             do j--; while (strcmp(Pair_Name[j],p)>0);
             if(j<i) break;
             s_swap2(Pair_Name,i,j);
             SWAP(brr[i],brr[j]);
          }
          strcpy(Pair_Name[m+1],Pair_Name[j]);
          strcpy(Pair_Name[j],p);
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}


/* creat an int matrix with subscript ange m[nrl...nrh][ncl...nch]  */
int     **imatrix(B64_long nrl,B64_long nrh,B64_long ncl,B64_long nch)
{
        B64_long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
        int  **m;

        /* allocate pointers to rows        */
        if((m=(int **)calloc(nrow,sizeof(int*)))==NULL)
        {
           printf("error imatrix: calloc error No. 1 \n");
           return(NULL);
        }
        m+=0;
        m-=nrl;

        /* allocate rows and set pointers to them        */
        if((m[nrl]=(int *)calloc(nrow*ncol,sizeof(int)))==NULL)
        {
           printf("error imatrix: calloc error No. 2 \n");
           return(NULL);
        }
        m[nrl]+=0;
        m[nrl]-=nrl;

        for(i=nrl+1;i<=nrh;i++)
           m[i]=m[i-1]+ncol;
        /* return pointer to array of pointers to rows   */
        return m;
}

/* creat char matrix with subscript ange cm[nrl...nrh][ncl...nch]  */
char    **cmatrix(B64_long nrl,B64_long nrh,B64_long ncl,B64_long nch)
{
        B64_long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
        char **cm;

        /* allocate pointers to rows        */
        if((cm=(char **)calloc(nrow,sizeof(char*)))==NULL)
        {
           printf("error cmatrix: calloc error No. 1 \n");
           return(NULL);
        }
        cm+=0;
        cm-=nrl;

        /* allocate rows and set pointers to them        */
        if((cm[nrl]=(char *)calloc(nrow*ncol,sizeof(char)))==NULL)
        {
           printf("error cmatrix: calloc error No. 2 \n");
           return(NULL);
        }
        cm[nrl]+=0;
        cm[nrl]-=nrl;

        for(i=nrl+1;i<=nrh;i++)
           cm[i]=cm[i-1]+ncol;
        /* return pointer to array of pointers to rows   */
        return cm;
}

